import json
from abc import ABC, abstractmethod
from os import path

from nile.signer import Signer
from pontis.core.const import NETWORK, ORACLE_CONTROLLER_ADDRESS
from starknet_py.contract import Contract, ContractData, ContractFunction
from starknet_py.net import Client
from starknet_py.net.models import InvokeFunction
from starkware.crypto.signature.signature import sign
from starkware.starknet.public.abi import get_selector_from_name

MAX_FEE = 0
FEE_SCALING_FACTOR = 1.1  # estimated fee is multiplied by this to set max_fee


class PontisBaseClient(ABC):
    def __init__(
        self,
        account_private_key,
        account_contract_address,
        network=None,
        oracle_controller_address=None,
        n_retries=None,
    ):
        if network is None:
            network = NETWORK
        if oracle_controller_address is None:
            oracle_controller_address = ORACLE_CONTROLLER_ADDRESS

        self.network = network
        self.oracle_controller_address = oracle_controller_address
        self.oracle_controller_contract = None
        self.account_contract_address = account_contract_address
        self.account_contract = None

        assert type(account_private_key) == int, "Account private key must be integer"
        self.account_private_key = account_private_key
        self.signer = Signer(self.account_private_key)

        self.client = Client(self.network, n_retries=n_retries)
        self.nonce = None

    @abstractmethod
    async def _fetch_contracts(self):
        pass

    async def get_nonce_uncached(self):
        await self._fetch_contracts()

        [nonce] = await self.client.call_contract(
            InvokeFunction(
                contract_address=self.account_contract_address,
                entry_point_selector=get_selector_from_name("get_nonce"),
                calldata=[],
                signature=[],
                max_fee=0,
                version=0,
            ),
            block_number="pending",
        )
        return nonce

    async def get_nonce(self):
        await self._fetch_contracts()

        nonce = await self.get_nonce_uncached()
        # If we have sent a tx recently, use local nonce because network state won't have been updated yet
        if self.nonce is not None and self.nonce >= nonce:
            nonce = self.nonce + 1

        self.nonce = nonce
        return nonce

    async def _fetch_base_contracts(self):
        if self.oracle_controller_contract is None:
            self.oracle_controller_contract = await Contract.from_address(
                self.oracle_controller_address,
                self.client,
            )

        if self.account_contract is None:
            self.account_contract = await Contract.from_address(
                self.account_contract_address, self.client
            )

    async def get_eth_balance(self):
        if self.network == "testnet":
            eth_address = (
                0x049D36570D4E46F48E99674BD3FCC84644DDD6B96F7C741B1562B82F9E004DC7
            )
        else:
            raise NotImplementedError(
                "PontisBaseClient.get_eth_balance: Unknown network type"
            )

        with open(path.join(path.dirname(__file__), "abi/ERC20.json"), "r") as f:
            erc20_abi = json.load(f)
        contract_data = ContractData.from_abi(eth_address, erc20_abi)
        balance_of_abi = [a for a in erc20_abi if a["name"] == "balanceOf"][0]
        balance_of_function = ContractFunction(
            "balanceOf", balance_of_abi, contract_data, self.client
        )

        result = await balance_of_function.call(self.account_contract_address)

        return result.balance

    async def send_transaction(
        self, to_contract, selector_name, calldata, max_fee=None
    ):
        return await self.send_transactions(
            [(to_contract, selector_name, calldata)], max_fee
        )

    async def send_transactions(self, calls, max_fee=None):
        nonce = await self.get_nonce()
        uncached_nonce = await self.get_nonce_uncached()

        # Format data for submission
        call_array = []
        offset = 0
        for i in range(len(calls)):
            call_array.append(
                {
                    "to": calls[i][0],
                    "selector": get_selector_from_name(calls[i][1]),
                    "data_offset": offset,
                    "data_len": len(calls[i][2]),
                }
            )
            offset += len(calls[i][2])
        calldata = [x for call in calls for x in call[2]]

        # Estimate fee
        with open(path.join(path.dirname(__file__), "abi/Account.json"), "r") as f:
            account_abi = json.load(f)
        contract_data = ContractData.from_abi(
            self.account_contract_address, account_abi
        )
        execute_abi = [a for a in account_abi if a["name"] == "__execute__"][0]
        execute_function = ContractFunction(
            "__execute__", execute_abi, contract_data, self.client
        )
        prepared = execute_function.prepare(
            call_array=call_array,
            calldata=calldata,
            nonce=uncached_nonce,  # have to use uncached because we call (not invoke), i.e. run against current starknet state
        )
        signature = sign(prepared.hash, self.account_private_key)
        # TODO: Change to using AccountClient once estimate_fee is fixed there
        tx = prepared._make_invoke_function(signature=signature)
        estimate = await prepared._client.estimate_fee(tx=tx)

        max_fee_estimate = int(estimate * FEE_SCALING_FACTOR)
        max_fee = (
            max_fee_estimate if max_fee is None else min(max_fee_estimate, max_fee)
        )

        # Submit transaction with fee
        prepared_with_fee = execute_function.prepare(
            call_array=call_array,
            calldata=calldata,
            nonce=nonce,
            max_fee=max_fee,
        )
        signature = sign(prepared_with_fee.hash, self.account_private_key)
        invocation = await prepared_with_fee.invoke(signature, max_fee=max_fee)

        return invocation



from .binance import fetch_binance
from .bitstamp import fetch_bitstamp
from .cex import fetch_cex
from .coinapi import fetch_coinapi
from .coinbase import fetch_coinbase
from .coingecko import fetch_coingecko
from .coinmarketcap import fetch_coinmarketcap
from .ftx import fetch_ftx
from .gemini import fetch_gemini



import os

import requests
from pontis.core.entry import construct_entry
from pontis.core.utils import currency_pair_to_key


def fetch_cex(assets):
    PUBLISHER_PREFIX = os.environ.get("PUBLISHER_PREFIX")
    publisher = PUBLISHER_PREFIX + "-cex"
    base_url = "https://cex.io/api/ticker"

    entries = []

    for asset in assets:
        if asset["type"] != "SPOT":
            print(f"Skipping CEX for non-spot asset {asset}")
            continue

        pair = asset["pair"]
        response = requests.get(f"{base_url}/{pair[0]}/{pair[1]}", timeout=10)
        result = response.json()

        if "error" in result and result["error"] == "Invalid Symbols Pair":
            print(f"No data found for {'/'.join(pair)} from CEX")
            continue

        timestamp = int(result["timestamp"])
        price = float(result["last"])
        price_int = int(price * (10 ** asset["decimals"]))
        key = currency_pair_to_key(*pair)

        print(f"Fetched price {price} for {'/'.join(pair)} from CEX")

        entries.append(
            construct_entry(
                key=key,
                value=price_int,
                timestamp=timestamp,
                publisher=publisher,
            )
        )

    return entries



from pontis.core.base_client import PontisBaseClient
from pontis.core.const import (
    ADMIN_ADDRESS,
    ORACLE_CONTROLLER_ADDRESS,
    PUBLISHER_REGISTRY_ADDRESS,
)
from pontis.core.utils import str_to_felt
from starknet_py.contract import Contract
from starknet_py.net import Client

MAX_FEE = 0
ADMIN_DEFAULT_N_RETRIES = 1


class PontisAdminClient(PontisBaseClient):
    def __init__(
        self,
        admin_private_key,
        admin_address=None,
        network=None,
        oracle_controller_address=None,
        publisher_registry_address=None,
        n_retries=None,
    ):
        if admin_address is None:
            admin_address = ADMIN_ADDRESS
        if publisher_registry_address is None:
            publisher_registry_address = PUBLISHER_REGISTRY_ADDRESS

        self.publisher_registry_address = publisher_registry_address
        self.publisher_registry_contract = None

        n_retries = ADMIN_DEFAULT_N_RETRIES if n_retries is None else n_retries
        super().__init__(
            admin_private_key,
            admin_address,
            network,
            oracle_controller_address,
            n_retries,
        )

    async def _fetch_contracts(self):
        await self._fetch_base_contracts()

        if self.publisher_registry_contract is None:
            self.publisher_registry_contract = await Contract.from_address(
                self.publisher_registry_address, Client(self.network)
            )

    async def get_primary_oracle_implementation_address(self):
        await self._fetch_contracts()

        result = await self.oracle_controller_contract.functions[
            "get_primary_oracle_implementation_address"
        ].call()

        return result.primary_oracle_implementation_address

    async def get_all_publishers(self):
        await self._fetch_contracts()

        result = await self.publisher_registry_contract.functions[
            "get_all_publishers"
        ].call()

        return result.publishers

    async def get_publisher_address(self, publisher):
        await self._fetch_contracts()

        result = await self.publisher_registry_contract.functions[
            "get_publisher_address"
        ].call(publisher)

        return result.publisher_address

    async def register_publisher_if_not_registered(self, publisher, publisher_address):
        if type(publisher) == str:
            publisher = str_to_felt(publisher)
        elif type(publisher) == int:
            publisher = publisher
        else:
            raise AssertionError(
                "Publisher ID must be string (will be converted to felt) or integer"
            )

        existing_publisher_address = await self.get_publisher_address(publisher)

        if existing_publisher_address == 0:
            result = await self.send_transaction(
                PUBLISHER_REGISTRY_ADDRESS,
                "register_publisher",
                [publisher, publisher_address],
            )
            print(f"Registered publisher with transaction {result}")

            return result

        else:
            print(
                f"Skipping registering {publisher}; already registered with address {existing_publisher_address}"
            )

        return

    async def add_oracle_implementation(self, oracle_implementation_address):
        result = await self.send_transaction(
            ORACLE_CONTROLLER_ADDRESS,
            "add_oracle_implementation_address",
            [oracle_implementation_address],
        )

        print(f"Added oracle implementation contract with transaction {result}")

        return result

    async def set_primary_oracle_implementation_address(
        self, primary_oracle_implementation_address
    ):
        result = await self.send_transaction(
            ORACLE_CONTROLLER_ADDRESS,
            "set_primary_oracle_implementation_address",
            [primary_oracle_implementation_address],
        )

        print(f"Set oracle implementation to primary with transaction {result}")

        return result

    async def update_oracle_implementation_active_status(
        self, oracle_implementation_address, is_active
    ):
        result = await self.send_transaction(
            ORACLE_CONTROLLER_ADDRESS,
            "update_oracle_implementation_active_status",
            [oracle_implementation_address, is_active],
        )

        print(f"Set active status on oracle implementation with transaction {result}")

        return result

    async def update_publisher_registry_address(self, new_publisher_registry_address):
        result = await self.send_transaction(
            ORACLE_CONTROLLER_ADDRESS,
            "update_publisher_registry_address",
            [new_publisher_registry_address],
        )

        print(f"Updated publisher registry address with transaction {result}")

        return result



import datetime
import os

import requests
from pontis.core.entry import construct_entry
from pontis.core.utils import currency_pair_to_key


def fetch_coinapi(assets):
    PUBLISHER_PREFIX = os.environ.get("PUBLISHER_PREFIX")
    publisher = PUBLISHER_PREFIX + "-coinapi"
    COINAPI_KEY = os.environ.get("COINAPI_KEY")
    headers = {"X-CoinAPI-Key": COINAPI_KEY}

    entries = []

    for asset in assets:
        if asset["type"] != "SPOT":
            print(f"Skipping Coinapi for non-spot asset {asset}")
            continue

        pair = asset["pair"]

        url = f"https://rest.coinapi.io/v1/exchangerate/{'/'.join(pair)}"
        response = requests.get(url, headers=headers, timeout=10)
        response.raise_for_status()
        price = response.json()["rate"]
        timestamp = int(
            datetime.datetime.strptime(
                response.json()["time"], "%Y-%m-%dT%H:%M:%S.%f0%z"
            ).timestamp()
        )
        price_int = int(price * (10 ** asset["decimals"]))
        key = currency_pair_to_key(*pair)

        print(f"Fetched price {price} for {key} from Coin API")

        entries.append(
            construct_entry(
                key=key,
                value=price_int,
                timestamp=timestamp,
                publisher=publisher,
            )
        )

    return entries



from pontis.core.base_client import PontisBaseClient
from pontis.core.entry import serialize_entries, serialize_entry

DEFAULT_N_RETRIES = 3


class PontisPublisherClient(PontisBaseClient):
    def __init__(
        self,
        publisher_private_key,
        publisher_address,
        network=None,
        oracle_controller_address=None,
        n_retries=None,
    ):
        n_retries = DEFAULT_N_RETRIES if n_retries is None else n_retries
        super().__init__(
            publisher_private_key,
            publisher_address,
            network,
            oracle_controller_address,
            n_retries,
        )

    async def _fetch_contracts(self):
        await self._fetch_base_contracts()

    async def publish(self, entry):
        result = await self.send_transaction(
            self.oracle_controller_address,
            "submit_entry",
            serialize_entry(entry),
        )
        print(f"Updated entry with transaction {result}")

        return result

    async def publish_many(self, entries):
        if len(entries) == 0:
            print("Skipping publishing as entries array is empty")
            return

        result = await self.send_transaction(
            self.oracle_controller_address,
            "submit_many_entries",
            serialize_entries(entries),
        )

        print(
            f"Successfully sent {len(entries)} updated entries  with transaction {result}"
        )

        return result



import os
import time

import requests
from pontis.core.entry import construct_entry
from pontis.core.utils import currency_pair_to_key


def fetch_gemini(assets):
    PUBLISHER_PREFIX = os.environ.get("PUBLISHER_PREFIX")
    publisher = PUBLISHER_PREFIX + "-gemini"

    base_url = "https://api.gemini.com/v1"
    response = requests.get(base_url + "/pricefeed", timeout=20)

    entries = []

    for asset in assets:
        if asset["type"] != "SPOT":
            print(f"Skipping Gemini for non-spot asset {asset}")
            continue

        pair = asset["pair"]
        key = currency_pair_to_key(*pair)
        timestamp = int(time.time())
        result = [e for e in response.json() if e["pair"] == "".join(pair)]
        if len(result) == 0:
            print(f"No entry found for {key} from Gemini")
            continue

        assert (
            len(result) == 1
        ), f"Found more than one matching entries for Gemini response and price pair {pair}"
        price = float(result[0]["price"])
        price_int = int(price * (10 ** asset["decimals"]))

        print(f"Fetched price {price} for {key} from Gemini")

        entries.append(
            construct_entry(
                key=key,
                value=price_int,
                timestamp=timestamp,
                publisher=publisher,
            )
        )

    return entries



PONTIS_ALL_ASSETS = [
    {"type": "SPOT", "pair": ("BTC", "USD")},
    {"type": "SPOT", "pair": ("ETH", "USD")},
    {"type": "SPOT", "pair": ("LUNA", "USD")},
    {"type": "SPOT", "pair": ("SOL", "USD")},
    {"type": "SPOT", "pair": ("AVAX", "USD")},
    {"type": "SPOT", "pair": ("DOGE", "USD")},
    {"type": "SPOT", "pair": ("SHIB", "USD")},
    {"type": "SPOT", "pair": ("TEMP", "USD")},
    {"type": "SPOT", "pair": ("DAI", "USD")},
    {"type": "SPOT", "pair": ("USDT", "USD")},
    {"type": "SPOT", "pair": ("USDC", "USD")},
    {"type": "SPOT", "pair": ("TUSD", "USD")},
    {"type": "SPOT", "pair": ("ETH", "MXN")},
    {"type": "FUTURE", "pair": ("BTC", "USD")},
    {"type": "FUTURE", "pair": ("ETH", "USD")},
]
